#include <EGL/egl.h>
#include <GLES2/gl2.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

//////////////////////////////////////////////////////////////////////////////
// static inlines...
#include "mat.c"
//////////////////////////////////////////////////////////////////////////////
#include "shader.c"

static const char *shad_color_vert =
"attribute vec4 vertexPosition;\n"
"attribute vec3 vertexColor;\n"
"varying vec3 fragColor;\n"
"uniform mat4 projection;\n"
"uniform mat4 modelView;\n"
"void main()\n"
"{\n"
"    gl_Position = projection * modelView * vertexPosition;\n"
"    fragColor = vertexColor;\n"
"}\n";

static const char *shad_color_frag =
"precision mediump float;\n"
"varying vec3 fragColor;\n"
"void main()\n"
"{\n"
"    gl_FragColor = vec4(fragColor, 1.0);\n"
"}\n";

static const char shad_tex_vert[] =
"attribute vec4 vertexPosition;\n"
"attribute vec2 vertexTextureCord;\n"
"varying vec2 textureCord;\n"
"uniform mat4 projection;\n"
"uniform mat4 modelView;\n"
"void main()\n"
"{\n"
"    gl_Position = projection * modelView * vertexPosition;\n"
"    textureCord = vertexTextureCord;\n"
"}\n";
static const char shad_tex_frag[] =
"precision mediump float;\n"
"uniform sampler2D texture;\n"
"uniform sampler2D texture_mask;\n"
"varying vec2 textureCord;\n"
"void main()\n"
"{\n"
"    float mask = texture2D(texture_mask, textureCord).r;\n"
"    if (mask < 0.5) discard;\n"
"    vec4 color = texture2D(texture, textureCord);\n"
"    gl_FragColor = vec4(color.r * mask, color.g * mask, color.b * mask, mask);\n"
"}\n";

// 72
static float cube_vert[] =
{
   -1,  1, -1,
    1,  1, -1,
   -1, -1, -1,
    1, -1, -1,
   -1,  1,  1,
    1,  1,  1,
   -1, -1,  1,
    1, -1,  1,
   -1,  1, -1,
   -1, -1, -1,
   -1, -1,  1,
   -1,  1,  1,
    1,  1, -1,
    1, -1, -1,
    1, -1,  1,
    1,  1,  1,
   -1, -1, -1,
   -1, -1,  1,
    1, -1,  1,
    1, -1, -1,
   -1,  1, -1,
   -1,  1,  1,
    1,  1,  1,
    1,  1, -1
};
// 72
#define COL1 0.5, 0.4, 0.3
#define COL2 0.3, 0.2, 0.1
#define COL3 0.4, 0.3, 0.2
#define COL4 0.1, 0.2, 0.3
#define COL5 0.2, 0.3, 0.4
#define COL6 0.4, 0.5, 0.6

static float cube_color[] =
{
   COL1, COL1, COL1, COL1,
   COL2, COL2, COL2, COL2,
   COL3, COL3, COL3, COL3,
   COL4, COL4, COL4, COL4,
   COL5, COL5, COL5, COL5,
   COL6, COL6, COL6, COL6,
};
// 72
static float cube_tex[] =
{
   1, 1,
   0, 1,
   1, 0,
   0, 0,

   0, 1,
   1, 1,
   0, 0,
   1, 0,

   0, 1,
   0, 0,
   1, 0,
   1, 1,

   1, 1,
   1, 0,
   0, 0,
   0, 1,

   0, 1,
   0, 0,
   1, 0,
   1, 1,

   0, 0,
   0, 1,
   1, 1,
   1, 0
};
// 36
static unsigned short cube_indices[] =
{
   0, 2, 3, 0, 1, 3, 4, 6, 7, 4, 5, 7, 8, 9, 10, 11, 8, 10, 12, 13,
   14, 15, 12, 14, 16, 17, 18, 16, 19, 18, 20, 21, 22, 20, 23, 22
};

#define WT 255, 255, 255, 255
#define BK   0,   0,   0, 255

#define YW 255, 255, 128, 255
#define YL 255, 255,   0, 255
#define OR 255, 128,   0, 255
#define RD 255,   0,   0, 255
#define RB 128,   0,   0, 255
#define BB   0,   0, 128, 255
#define BL   0,   0, 255, 255
#define BC   0, 128, 255, 255
#define CY   0, 255, 255, 255
#define CW 128, 255, 255, 255

// 8x8 @ 32bpp
static unsigned char tex_color[] =
{
   WT, WT, YW, YL, OR, RD, RB, BK,
   WT, YW, YL, OR, RD, RB, BK, BB,
   YW, YL, OR, RD, RB, BK, BB, BL,
   YL, OR, RD, RB, BK, BB, BL, BC,
   OR, RD, RB, BK, BB, BL, BC, CY,
   RD, RB, BK, BB, BL, BC, CY, CW,
   RB, BK, BB, BL, BC, CY, CW, WT,
   BK, BB, BL, BC, CY, CW, WT, WT
};

#define OOO 255
#define _I_   0

// 8x8 @ 8bpp
static unsigned char tex_mask[] =
{
   _I_, _I_, OOO, OOO, OOO, OOO, _I_, _I_,
   _I_, OOO, OOO, OOO, OOO, OOO, OOO, _I_,
   OOO, OOO, OOO, _I_, _I_, OOO, OOO, OOO,
   OOO, OOO, _I_, _I_, _I_, _I_, OOO, OOO,
   OOO, OOO, _I_, _I_, _I_, _I_, OOO, OOO,
   OOO, OOO, OOO, _I_, _I_, OOO, OOO, OOO,
   _I_, OOO, OOO, OOO, OOO, OOO, OOO, _I_,
   _I_, _I_, OOO, OOO, OOO, OOO, _I_, _I_
};

//////////////////////////////////////////////////////////////////////////////

typedef struct
{
   unsigned int prog;
   struct {
      unsigned int vert;
      unsigned int color;
      unsigned int tex;
   } attrib;
   struct {
      unsigned int projection;
      unsigned int model_view;
   } uniform;
   struct {
      unsigned int color;
      unsigned int mask;
   } texture;
   int points;
   float *vert;
   float *color;
   float *tex;
   int indexes;
   unsigned short *index;
} Obj;

static float matr_projection[16];
static Obj *cube1 = NULL;
static Obj *cube2 = NULL;

static unsigned int
tex_new(unsigned char *pix, int bpp, int w, int h)
{
   unsigned int tex;

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glGenTextures(1, &tex);
   if (bpp == 32)
     {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, tex);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, pix);
     }
   else
     {
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, tex);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, w, h, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, pix);
     }
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   return tex;
}

static Obj *
obj_new(const char *vert_shader, const char *frag_shader,
        int points, float *vert, float *color, float *tex,
        int indexes, unsigned short *index,
        unsigned char *tex_col, int tex_col_w, int tex_col_h,
        unsigned char *mask_col, int tex_mask_w, int tex_mask_h
       )
{
   Obj *o = calloc(1, sizeof(Obj));
   if (!o) abort();
   o->prog = shad_prog_new(vert_shader, frag_shader);
   o->uniform.projection = glGetUniformLocation(o->prog, "projection");
   o->uniform.model_view = glGetUniformLocation(o->prog, "modelView");

   o->points = points;
   o->indexes = indexes;

   o->vert  = calloc(o->points * 3, sizeof(float));
   o->attrib.vert = glGetAttribLocation(o->prog, "vertexPosition");
   memcpy(o->vert, vert, o->points * 3 * sizeof(float));

   if (color)
     {
        o->attrib.color = glGetAttribLocation(o->prog, "vertexColor");
        o->color = calloc(o->points * 3, sizeof(float));
        memcpy(o->color, color, o->points * 3 * sizeof(float));
     }
   if (tex)
     {
        o->attrib.tex = glGetAttribLocation(o->prog, "vertexTextureCord");
        o->tex = calloc(o->points * 2, sizeof(float));
        memcpy(o->tex, tex, o->points * 2 * sizeof(float));
     }

   o->index = calloc(o->indexes, sizeof(unsigned short));
   memcpy(o->index, index, o->indexes * sizeof(unsigned short));

   if (tex_col)
     o->texture.color = tex_new(tex_col, 32, tex_col_w, tex_col_h);
   if (tex_mask)
     o->texture.mask  = tex_new(tex_mask,  8, tex_mask_w, tex_mask_h);
   return o;
}

void
obj_draw(Obj *o, float *modview)
{
//   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//   glEnable(GL_BLEND);
   glDisable(GL_BLEND);

   glUseProgram(o->prog);
   glUniform1i(glGetUniformLocation(o->prog, "texture"), 0);
   glUniform1i(glGetUniformLocation(o->prog, "texture_mask"), 1);
   glVertexAttribPointer(o->attrib.vert, 3, GL_FLOAT, GL_FALSE, 0, o->vert);
   glEnableVertexAttribArray(o->attrib.vert);
   if (o->color)
     {
        glVertexAttribPointer(o->attrib.color, 3, GL_FLOAT, GL_FALSE, 0, o->color);
        glEnableVertexAttribArray(o->attrib.color);
     }
   if ((o->texture.color) && (o->tex))
     {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, o->texture.color);
        glVertexAttribPointer(o->attrib.tex, 2, GL_FLOAT, GL_FALSE, 0, o->tex);
        glEnableVertexAttribArray(o->attrib.tex);
     }
   if ((o->texture.mask) && (o->tex))
     {
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, o->texture.mask);
        glVertexAttribPointer(o->attrib.tex, 2, GL_FLOAT, GL_FALSE, 0, o->tex);
        glEnableVertexAttribArray(o->attrib.tex);
     }
   glUniformMatrix4fv(o->uniform.projection, 1, GL_FALSE, matr_projection);
   glUniformMatrix4fv(o->uniform.model_view, 1, GL_FALSE, modview);
   glDrawElements(GL_TRIANGLES, o->indexes, GL_UNSIGNED_SHORT, o->index);
}

void
draw_init(int w, int h)
{
   cube1 = obj_new(shad_color_vert, shad_color_frag,
                   24, cube_vert, cube_color, NULL,
                   36, cube_indices,
                   NULL, 0, 0,
                   NULL, 0, 0
                  );

   cube2 = obj_new(shad_tex_vert, shad_tex_frag,
                   24, cube_vert, NULL, cube_tex,
                   36, cube_indices,
                   tex_color, 8, 8,
                   tex_mask, 8, 8
                  );

   mat_persp(matr_projection, 45, (float)w / (float)h, 0.1, 100);

   glEnable(GL_DEPTH_TEST);
   glViewport(0, 0, w, h);
}

void
draw(int w, int h)
{
   static double ang = 0.0;
   float modview[16];

   glClearColor(0.25, 0.25, 0.25, 1.0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

   mat_ident(modview);
   mat_rot_x(modview, (ang * 2.0));
   mat_rot_y(modview, (ang * 2.0) + 23);
   mat_trans(modview, 0, 0, -10);
   obj_draw(cube2, modview);

   mat_ident(modview);
   mat_rot_x(modview, ang + 71);
   mat_rot_y(modview, ang + 23);
   mat_scale(modview, 10, 10, 10);
   mat_trans(modview, 0, 0, -10);
   obj_draw(cube1, modview);

   // rotate  + loop
   ang += 0.1;
   if (ang > 360) ang -= 360;
}

#if 0
GLuint loadSimpleTexture()
{
       /* Texture Object Handle. */
       GLuint textureId;
       /* 3 x 3 Image,  R G B A Channels RAW Format. */
       GLubyte pixels[9 * 4] =
     {
                    18,  140, 171, 255, /* Some Colour Bottom Left. */
                    143, 143, 143, 255, /* Some Colour Bottom Middle. */
                    255, 255, 255, 255, /* Some Colour Bottom Right. */
                    255, 255, 0,   255, /* Yellow Middle Left. */
                    0,   255, 255, 255, /* Some Colour Middle. */
                    255, 0,   255, 255, /* Some Colour Middle Right. */
                    255, 0,   0,   255, /* Red Top Left. */
                    0,   255, 0,   255, /* Green Top Middle. */
                    0,   0,   255, 255, /* Blue Top Right. */
     };
       /* Use tightly packed data. */
       glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
       /* Generate a texture object. */
       glGenTextures(1, &textureId);
       /* Activate a texture. */
       glActiveTexture(GL_TEXTURE0);
       /* Bind the texture object. */
       glBindTexture(GL_TEXTURE_2D, textureId);
       /* Load the texture. */
       glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 3, 3, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
       /* Set the filtering mode. */
       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
       return textureId;
}

static const char glVertexShader[] =
        "attribute vec4 vertexPosition;\n"
        "attribute vec2 vertexTextureCord;\n"
        "varying vec2 textureCord;\n"
        "uniform mat4 projection;\n"
        "uniform mat4 modelView;\n"
        "void main()\n"
        "{\n"
        "    gl_Position = projection * modelView * vertexPosition;\n"
        "    textureCord = vertexTextureCord;\n"
        "}\n";
static const char glFragmentShader[] =
        "precision mediump float;\n"
        "uniform sampler2D texture;\n"
        "varying vec2 textureCord;\n"
        "void main()\n"
        "{\n"
        "    gl_FragColor = texture2D(texture, textureCord);\n"
        "}\n";
bool setupGraphics(int width, int height)
{
       glProgram = createProgram(glVertexShader, glFragmentShader);
       if (!glProgram)
     {
                LOGE ("Could not create program");
                return false;
     }
       vertexLocation = glGetAttribLocation(glProgram, "vertexPosition");
       textureCordLocation = glGetAttribLocation(glProgram, "vertexTextureCord");
       projectionLocation = glGetUniformLocation(glProgram, "projection");
       modelViewLocation = glGetUniformLocation(glProgram, "modelView");
       samplerLocation = glGetUniformLocation(glProgram, "texture");
       /* Setup the perspective. */
       matrixPerspective(projectionMatrix, 45, (float)width / (float)height, 0.1f, 100);
       glEnable(GL_DEPTH_TEST);
       glViewport(0, 0, width, height);
       /* Load the Texture. */
       textureId = loadSimpleTexture();
       if(textureId == 0)
     {
                return false;
     }
       else
     {
                return true;
     }
}
GLfloat cubeVertices[] = {-1.0f,  1.0f, -1.0f, /* Back. */
                              1.0f,  1.0f, -1.0f,
                             -1.0f, -1.0f, -1.0f,
                              1.0f, -1.0f, -1.0f,
                             -1.0f,  1.0f,  1.0f, /* Front. */
                              1.0f,  1.0f,  1.0f,
                             -1.0f, -1.0f,  1.0f,
                              1.0f, -1.0f,  1.0f,
                             -1.0f,  1.0f, -1.0f, /* Left. */
                             -1.0f, -1.0f, -1.0f,
                             -1.0f, -1.0f,  1.0f,
                             -1.0f,  1.0f,  1.0f,
                              1.0f,  1.0f, -1.0f, /* Right. */
                              1.0f, -1.0f, -1.0f,
                              1.0f, -1.0f,  1.0f,
                              1.0f,  1.0f,  1.0f,
                             -1.0f, 1.0f, -1.0f, /* Top. */
                             -1.0f, 1.0f,  1.0f,
                              1.0f, 1.0f,  1.0f,
                              1.0f, 1.0f, -1.0f,
                             -1.0f, - 1.0f, -1.0f, /* Bottom. */
                             -1.0f,  -1.0f,  1.0f,
                              1.0f, - 1.0f,  1.0f,
                              1.0f,  -1.0f, -1.0f
};
GLfloat textureCords[] = { 1.0f, 1.0f, /* Back. */
                              0.0f, 1.0f,
                              1.0f, 0.0f,
                              0.0f, 0.0f,
                              0.0f, 1.0f, /* Front. */
                              1.0f, 1.0f,
                              0.0f, 0.0f,
                              1.0f, 0.0f,
                              0.0f, 1.0f, /* Left. */
                              0.0f, 0.0f,
                              1.0f, 0.0f,
                              1.0f, 1.0f,
                              1.0f, 1.0f, /* Right. */
                              1.0f, 0.0f,
                              0.0f, 0.0f,
                              0.0f, 1.0f,
                              0.0f, 1.0f, /* Top. */
                              0.0f, 0.0f,
                              1.0f, 0.0f,
                              1.0f, 1.0f,
                              0.0f, 0.0f, /* Bottom. */
                              0.0f, 1.0f,
                              1.0f, 1.0f,
                              1.0f, 0.0f
};

    glVertexAttribPointer(textureCordLocation, 2, GL_FLOAT, GL_FALSE, 0, textureCords);
    glEnableVertexAttribArray(textureCordLocation);
    glUniformMatrix4fv(projectionLocation, 1, GL_FALSE,projectionMatrix);
    glUniformMatrix4fv(modelViewLocation, 1, GL_FALSE, modelViewMatrix);
    /* Set the sampler texture unit to 0. */
    glUniform1i(samplerLocation, 0);
#endif
